"""
!/usr/bin/env python3
-*- coding: utf-8 -*-

----------------------------------------------------------------------
Copyright 2021 CFS Engineering

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
----------------------------------------------------------------------

Author: Aidan Jungo

"""

from pathlib import Path


# PATH
CPACSPY_LIB = Path(__file__).absolute().parent
CPACSPY_ROOT = CPACSPY_LIB.parents[1]
EXAMPLES_PATH = Path(CPACSPY_ROOT, "examples")
TESTS_PATH = Path(CPACSPY_ROOT, "tests")
D150_TESTS_PATH = Path(TESTS_PATH, "D150_simple.xml")
PROPELLER_TESTS_PATH = Path(TESTS_PATH, "SimpleAircraft_propeller.xml")

# XPATH
AC_NAME_XPATH = "/cpacs/header/name"
AIRCRAFT_XPATH = "/cpacs/vehicles/aircraft/model"
ROTORCRAFT_XPATH = "/cpacs/vehicles/rotorcraft/model"
AEROPERFORMANCE_XPATH = "/cpacs/vehicles/aircraft/model/analyses/aeroPerformance"

# Lists
PARAMS = ["altitude", "machNumber", "angleOfSideslip", "angleOfAttack"]
COEFS = ["cd", "cl", "cs", "cmd", "cml", "cms"]
PARAMS_COEFS = PARAMS + COEFS
DAMPING_COEFS = [
    "dcddpStar",
    "dcddqStar",
    "dcddrStar",
    "dcldpStar",
    "dcldqStar",
    "dcldrStar",
    "dcmddpStar",
    "dcmddqStar",
    "dcmddrStar",
    "dcmldpStar",
    "dcmldqStar",
    "dcmldrStar",
    "dcmsdpStar",
    "dcmsdqStar",
    "dcmsdrStar",
    "dcsdpStar",
    "dcsdqStar",
    "dcsdrStar",
]


# Stability messages
MSG_STAB_ONE_PARAM = "Warning, normally stability should be check for one parameter at the time."
MSG_STAB_NOT_ENOUGH = "Not enough value to check stability"
MSG_STAB_NEUTRAL = "Neutral stability"


def listify(value):
    """If variable, return a list of 1 value, if already a list don't change a list."""

    if value is not None:
        if not isinstance(value, list):
            value = [value]
    else:
        value = []

    return value



"""
!/usr/bin/env python3
-*- coding: utf-8 -*-

----------------------------------------------------------------------
Copyright 2021 CFS Engineering

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
----------------------------------------------------------------------

Author: Aidan Jungo

"""

from pathlib import Path

import numpy as np
import pandas as pd

from cpacspy.aeromap import AeroMap
from cpacspy.aircraft import Aircraft
from cpacspy.cpacsfunctions import get_xpath_parent, open_tigl, open_tixi
from cpacspy.rotorcraft import Rotorcraft
from cpacspy.utils import AC_NAME_XPATH, AEROPERFORMANCE_XPATH, AIRCRAFT_XPATH, ROTORCRAFT_XPATH


class CPACS:
    """CPACS class"""

    def __init__(self, cpacs_file):

        # To accept either a Path or a string
        if isinstance(cpacs_file, Path):
            self.cpacs_file = str(cpacs_file)
        else:
            self.cpacs_file = cpacs_file

        # CPACS
        self.tixi = open_tixi(cpacs_file)
        self.tigl = open_tigl(self.tixi)

        # Aircraft name
        if self.tixi.checkElement(AC_NAME_XPATH):
            self.ac_name = self.tixi.getTextElement(AC_NAME_XPATH)

        # Aircraft data
        if self.tixi.checkElement(AIRCRAFT_XPATH):
            self.aircraft = Aircraft(self.tixi, self.tigl)

        # Rotorcraft data
        if self.tixi.checkElement(ROTORCRAFT_XPATH):
            self.tigl_rotor = open_tigl(self.tixi, rotorcraft=True)
            self.rotorcraft = Rotorcraft(self.tixi, self.tigl_rotor)

        # Load aeroMaps
        self.load_all_aeromaps()

    def load_all_aeromaps(self):
        """Load all the aeromaps present in the CPACS file as object."""

        self.nb_aeromaps = 0
        self.aeromaps = []

        for aeromap_uid in self.get_aeromap_uid_list():
            aeromap = AeroMap(self.tixi, aeromap_uid)
            self.aeromaps.append(aeromap)
            self.nb_aeromaps += 1

    def get_aeromap_uid_list(self):
        """Get the list of all aeroMap UID."""

        uid_list = []

        if not self.tixi.checkElement(AEROPERFORMANCE_XPATH):
            return uid_list

        aeromap_count = self.tixi.getNamedChildrenCount(AEROPERFORMANCE_XPATH, "aeroMap")
        if aeromap_count:
            for i in range(aeromap_count):
                aeromap_xpath = AEROPERFORMANCE_XPATH + "/aeroMap[" + str(i + 1) + "]"
                aeromap_uid = self.tixi.getTextAttribute(aeromap_xpath, "uID")
                uid_list.append(aeromap_uid)
        else:
            print('No "aeroMap" has been found in this CPACS file')

        return uid_list

    def get_aeromap_by_uid(self, uid):
        """Get an aeromap object by its uid."""

        for aeromap in self.aeromaps:
            if aeromap.uid == uid:
                return aeromap

        raise ValueError(f'No aeromap with "{uid}" as uid as been found!')

    def create_aeromap(self, uid):
        """Create a new aeromap object."""

        if " " in uid:
            raise ValueError("AeroMap uid should not contain any space!")

        if uid not in self.get_aeromap_uid_list():
            new_aeromap = AeroMap(self.tixi, uid, create_new=True)
            self.aeromaps.append(new_aeromap)
            self.nb_aeromaps += 1
            return new_aeromap
        else:
            raise ValueError("This uid already exit!")

    def create_aeromap_from_csv(self, csv_path, uid=None):
        """Create a new aeromap object from a CSV file."""

        if isinstance(csv_path, str):
            csv_path = Path(csv_path)

        if not uid:
            uid = csv_path.stem

        if not csv_path.exists():
            raise ValueError(f"CSV file not found at {csv_path.absolute}")

        new_aeromap = self.create_aeromap(uid)

        new_aeromap.df = pd.read_csv(
            csv_path, keep_default_na=True, na_values=float(np.nan), dtype=float
        )

        return new_aeromap

    def duplicate_aeromap(self, uid_base, uid_duplicate):
        """Duplicate an aeromap and return the new aeromap object."""

        # Check uid's
        if uid_base not in self.get_aeromap_uid_list():
            raise ValueError("The AeroMap to duplicate does not exit!")

        if uid_duplicate in self.get_aeromap_uid_list():
            raise ValueError("This uid for the duplicate already exit!")

        # Get AeroMap and duplicate
        am_base = self.get_aeromap_by_uid(uid_base)
        am_duplicated = AeroMap(self.tixi, uid_duplicate, create_new=True)

        # Copy data
        am_duplicated.df = am_base.df
        am_duplicated.description = am_base.description + f' (duplicate from "{uid_base}")'

        self.aeromaps.append(am_duplicated)
        self.nb_aeromaps += 1

        return am_duplicated

    def delete_aeromap(self, uid):
        """Delete an aeromap from its uid."""

        # Check if uid is valid
        if " " in uid:
            raise ValueError("AeroMap uid should not contain any space!")

        if uid not in self.get_aeromap_uid_list():
            raise ValueError(f'uid "{uid}"" does not exit! The aeroMap cannot be deleted!')

        # Remove the aeromap  from the CPACS file
        aeromap = self.get_aeromap_by_uid(uid)
        xpath = get_xpath_parent(aeromap.xpath, level=1)
        self.tixi.removeElement(xpath)

        # Reload the aeromaps to take into account the changes in the CPACS file
        self.load_all_aeromaps()

    def save_cpacs(self, cpacs_file, overwrite=False):
        """Save a CPACS file from the TIXI object at a chosen path."""

        # To accept either a Path or a string
        if isinstance(cpacs_file, str):
            cpacs_file = Path(cpacs_file)

        # Check for .xml file
        if cpacs_file.suffix != ".xml":
            raise ValueError("The CPACS file name must be a .xml file!")

        # Check if file name must be change to avoid overwrite
        if cpacs_file.exists() and not overwrite:
            find_name = False
            i = 1
            while not find_name:
                cpacs_file_new_name = Path(cpacs_file.parent, f"{cpacs_file.stem}_{i}.xml")
                if not cpacs_file_new_name.exists():
                    find_name = True
                    cpacs_file = cpacs_file_new_name
                else:
                    i += 1

        self.tixi.save(str(cpacs_file))

    def __str__(self):

        text_line = []
        text_line.append("\nCPACS file ----------------------------------------------------------")
        text_line.append(" ")
        text_line.append(f"Aircraft name : {self.ac_name}")
        text_line.append(f"CPACS file path: {self.cpacs_file}")
        text_line.append(" ")
        text_line.append("List of AeroMaps:")
        for aeromap_uid in self.get_aeromap_uid_list():
            text_line.append("  " + aeromap_uid)
        text_line.append(" ")
        text_line.append("---------------------------------------------------------------------\n")

        return ("\n").join(text_line)


if __name__ == "__main__":
    pass



"""
!/usr/bin/env python3
-*- coding: utf-8 -*-

----------------------------------------------------------------------
Copyright 2021 CFS Engineering

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
----------------------------------------------------------------------

Author: Aidan Jungo

"""

import math

import numpy as np
import pandas as pd
from ambiance import Atmosphere
from matplotlib import pyplot as plt
from scipy import stats

from cpacspy.cpacsfunctions import (
    add_float_vector,
    create_branch,
    get_float_vector,
    get_xpath_parent,
)
from cpacspy.utils import (
    AEROPERFORMANCE_XPATH,
    COEFS,
    DAMPING_COEFS,
    PARAMS,
    PARAMS_COEFS,
    listify,
)

from cpacspy.utils import MSG_STAB_NEUTRAL, MSG_STAB_NOT_ENOUGH, MSG_STAB_ONE_PARAM


def get_filter(df, alt_list, mach_list, aos_list, aoa_list):
    """Get a dataframe filter for a set of parameters lists."""

    filter = pd.Series(True, index=df.index)

    if alt_list:
        filter &= df["altitude"].isin(alt_list)

    if mach_list:
        filter &= df["machNumber"].isin(mach_list)

    if aos_list:
        filter &= df["angleOfSideslip"].isin(aos_list)

    if aoa_list:
        filter &= df["angleOfAttack"].isin(aoa_list)

    return filter


class AeroMap:
    """AeroMap class for CPACS AeroMap."""

    def __init__(self, tixi, uid, create_new=False):
        """Init aeromap class

        Args:
            tixi (object): TIXI object open from a CPACS file
            uid (str): UID of the AeroMap
            create_new (bool, optional): If True crate a new AeroMap in TIXI,
                                         if False find it in the CPACS file.
                                         Defaults to False.
        """

        self.tixi = tixi
        self.uid = uid
        self.name = uid
        self.description = ""
        self.atmospheric_model = "ISA"
        self.df = pd.DataFrame(columns=PARAMS_COEFS, dtype=float)

        if create_new:
            self.name = uid
            self.xpath = None

        else:
            self.xpath = self.tixi.uIDGetXPath(uid) + "/aeroPerformanceMap"
            name_xpath = get_xpath_parent(self.xpath) + "/name"
            if self.tixi.checkElement(name_xpath):
                self.name = self.tixi.getTextElement(name_xpath)

            description_xpath = get_xpath_parent(self.xpath) + "/description"
            if self.tixi.checkElement(description_xpath):
                self.description = self.tixi.getTextElement(description_xpath)

            atm_model_xpath = get_xpath_parent(self.xpath) + "/boundaryConditions/atmosphericModel"
            if self.tixi.checkElement(atm_model_xpath):
                self.atmospheric_model = self.tixi.getTextElement(atm_model_xpath)

            self.get_param_and_coef_from_cpacs()

    def get_param_and_coef_from_cpacs(self):
        """Get the parameters and coefficients from the aeroMap of a CPACS file."""

        param_dict = {}

        # Get parameters
        for param in PARAMS:
            param_xpath = self.xpath + f"/{param}"

            if self.tixi.checkElement(param_xpath):
                param_dict[param] = get_float_vector(self.tixi, param_xpath)
            else:
                raise ValueError(f'No value has been found for "{param}" in "{self.uid}" aeroMap!')

        # Get coefficients
        for coef in COEFS:
            coef_xpath = self.xpath + f"/{coef}"

            if self.tixi.checkElement(coef_xpath):
                param_dict[coef] = get_float_vector(self.tixi, coef_xpath)

        # Get damping derivatives coefficients
        for rates in ["negativeRates", "positiveRates"]:
            for damping_coef in DAMPING_COEFS:

                col_name = f"dampingDerivatives_{rates}_{damping_coef}"
                coef_xpath = self.xpath + f"/dampingDerivatives/{rates}/{damping_coef}"

                if self.tixi.checkElement(coef_xpath):
                    param_dict[col_name] = get_float_vector(self.tixi, coef_xpath)

        # Get incrementMaps coefficients
        # TODO

        df_param = pd.DataFrame(param_dict)
        self.df = pd.concat([self.df, df_param], axis=0)

    def get(self, list_of, alt=None, mach=None, aos=None, aoa=None):
        """Get parameter or coeffs as a numpy vector with other parameters as filter (optional).

        Args:
            list_of (str): Parameter or coefficient to get.
            alt (list, optional): List of altitudes to filter. Defaults to None.
            mach (list, optional): List of Mach numbers to filter. Defaults to None.
            aos (list, optional): List of angle of sideslip to filter. Defaults to None.
            aoa (list, optional): List of angle of attack to filter. Defaults to None.

        """

        alt_list = listify(alt)
        mach_list = listify(mach)
        aos_list = listify(aos)
        aoa_list = listify(aoa)

        filt = get_filter(self.df, alt_list, mach_list, aos_list, aoa_list)

        return self.df.loc[filt, list_of].to_numpy()

    def get_damping_derivatives(self, coef, axis, rates, alt=None, mach=None, aos=None, aoa=None):
        """Get damping derivatives coefficients as a numpy vector with other parameters as
        filter (optional).

        Args:
            coef (str): Coefficient to get ['cl','cd','cm','cml','cmd','cms']
            axis (str): Axis to get ['dp','dq','dr']
            rates (str): Rates to get ['pos','neg']
            alt (float, optional): Altitude
            mach (float, optional): Mach number
            aos (float, optional): Angle of sideslip
            aoa (float, optional): Angle of attack

        """

        if coef not in COEFS:
            raise ValueError(
                f'{coef} is not a valid coefficient! \n \
                Must be one of the following: {",".join(COEFS)}'
            )

        if axis not in ["dp", "dq", "dr"]:
            raise ValueError(
                f'{axis} is not a valid axis! \n \
                Must be one of the following: {"dp","dq","dr"}'
            )

        if rates in ["positive", "pos", "p"]:
            rates_name = "positiveRates"
        elif rates in ["negative", "neg", "n"]:
            rates_name = "negativeRates"
        else:
            raise ValueError(
                f'Invalid rates "{rates}"! \n \
                             Must be written as: \n \
                                 ("positive","pos","p" or "+") for positiveRates or \n \
                                     ("negative","neg","n" or "-") for negativeRates'
            )

        col_name = f"dampingDerivatives_{rates_name}_d{coef}{axis}Star"

        if col_name not in self.df.columns:
            raise ValueError(
                f"Damping derivative for {coef} coefficient on {axis} axis \
                with {rates_name} has not been found!"
            )

        return self.get(col_name, alt=alt, mach=mach, aos=aos, aoa=aoa)

    def add_row(
        self,
        alt,
        mach,
        aos,
        aoa,
        cd=np.nan,
        cl=np.nan,
        cs=np.nan,
        cmd=np.nan,
        cml=np.nan,
        cms=np.nan,
    ):
        """Add a row in an Aeromap dataframe.

        Args:
            alt (float): Altitude
            mach (float): Mach number
            aos (float): Angle of sideslip
            aoa (float): Angle of attack
            cd (float, optional): Coefficient of drag
            cl (float, optional): Coefficient of lift
            cs (float, optional): Coefficient of side force
            cmd (float, optional): Coefficient of moment due to drag
            cml (float, optional): Coefficient of moment due to lift
            cms (float, optional): Coefficient of moment due to side force

        """

        # Check if the parameter already exists
        filt = get_filter(self.df, [alt], [mach], [aos], [aoa])
        if not self.df.loc[filt].empty:
            raise ValueError(
                f"Row with alt={alt}, mach={mach}, aos={aos}, aoa={aoa} already exists!"
            )

        new_row = {
            "altitude": alt,
            "machNumber": mach,
            "angleOfSideslip": aos,
            "angleOfAttack": aoa,
            "cd": cd,
            "cl": cl,
            "cs": cs,
            "cmd": cmd,
            "cml": cml,
            "cms": cms,
        }

        # Check if all column already exist
        for col in new_row:
            if col not in self.df.columns:
                self.df[col] = np.nan

        # Add the new row
        self.df = self.df.append(new_row, ignore_index=True)

    def remove_row(self, alt, mach, aos, aoa):
        """Remove a row in an Aeromap dataframe for a set of parameters.

        Args:
            alt (float): Altitude
            mach (float): Mach number
            aos (float): Angle of sideslip
            aoa (float): Angle of attack

        """

        # Check if the parameter exists
        filt = get_filter(self.df, [alt], [mach], [aos], [aoa])

        if filt.empty or filt.any() == False:
            raise ValueError(
                f'No values has been found for \
                {alt}, {mach}, {aos}, {aoa} in "{self.uid}" aeroMap!'
            )

        # Remove the row
        self.df = self.df.drop(self.df.loc[filt].index)

    def add_coefficients(
        self,
        alt,
        mach,
        aos,
        aoa,
        cd=np.nan,
        cl=np.nan,
        cs=np.nan,
        cmd=np.nan,
        cml=np.nan,
        cms=np.nan,
    ):
        """Add coefficients to existing set of parameter.

        Args:
            alt (float): Altitude
            mach (float): Mach number
            aos (float): Angle of sideslip
            aoa (float): Angle of attack
            cd (float, optional): Coefficient of drag
            cl (float, optional): Coefficient of lift
            cs (float, optional): Coefficient of side force
            cmd (float, optional): Coefficient of moment due to drag
            cml (float, optional): Coefficient of moment due to lift
            cms (float, optional): Coefficient of moment due to side force

        """

        # Check if parameter are already in the dataframe
        filt = get_filter(self.df, [alt], [mach], [aos], [aoa])

        if filt.empty or filt.any() == False:
            raise ValueError(
                f'No values has been found for \
                {alt}, {mach}, {aos}, {aoa} in "{self.uid}" aeroMap!'
            )

        self.df.loc[filt, ["cd", "cl", "cs", "cmd", "cml", "cms"]] = [cd, cl, cs, cmd, cml, cms]

    def add_damping_derivatives(self, alt, mach, aos, aoa, coef, axis, value, rate=-1.0):
        """Add a damping derivative coefficients for an existing set of alt,mach,aos,aoa.

        Args:
            alt (float): Altitude
            mach (float): Mach number
            aos (float): Angle of sideslip
            aoa (float): Angle of attack
            coef (str) : equivalent aero coefficient (cd,cl,cs,cmd,cml,cms)
            axis (str) : axis of rotation (dp,dq,dr)
            value (float) : value of the coefficient
            rate (float, optional): Rotation rate. Defaults to -1.0.

        """

        damping_coef = "d" + coef + axis + "Star"

        if damping_coef not in DAMPING_COEFS:
            raise ValueError(f'Damping coefficient "{damping_coef}" is not valid!')

        if rate < 0:
            col_name = "dampingDerivatives_negativeRates_" + damping_coef
        elif rate > 0:
            col_name = "dampingDerivatives_positiveRates_" + damping_coef
        else:
            raise ValueError(f'Rotation rate "{rate}" is not valid!')

        # Check if this set of parameters exists
        filter = get_filter(self.df, [alt], [mach], [aos], [aoa])
        if not filter.any():
            raise ValueError(
                f"Parameters for alt={alt}, mach={mach}, aos={aos}, aoa={aoa} \
                do not exist!"
            )

        self.df.loc[
            (self.df["altitude"] == alt)
            & (self.df["machNumber"] == mach)
            & (self.df["angleOfSideslip"] == aos)
            & (self.df["angleOfAttack"] == aoa),
            [col_name],
        ] = value

    def plot(self, x_param, y_param, alt=None, mach=None, aos=None, aoa=None):
        """Plot 'x_param' vs 'y_param' with filtered parameters passed as float or string."""

        alt_list = listify(alt)
        mach_list = listify(mach)
        aos_list = listify(aos)
        aoa_list = listify(aoa)

        filt = get_filter(self.df, alt_list, mach_list, aos_list, aoa_list)
        self.df.loc[filt].plot(x=x_param, y=y_param, ylabel=y_param, legend=False, marker="o")
        plt.show()

    def save(self):
        """Save the AeroMap in the TIXI object."""

        # Create and fill the '/aeroPerformanceMap' field
        if not self.xpath:
            if self.tixi.checkElement(AEROPERFORMANCE_XPATH):
                child_nb = self.tixi.getNumberOfChilds(AEROPERFORMANCE_XPATH)
                self.tixi.createElementAtIndex(AEROPERFORMANCE_XPATH, "aeroMap", child_nb + 1)
                self.xpath = AEROPERFORMANCE_XPATH + f"/aeroMap[{child_nb+1}]/aeroPerformanceMap"
            else:
                create_branch(self.tixi, AEROPERFORMANCE_XPATH + "/aeroMap")
                self.xpath = AEROPERFORMANCE_XPATH + "/aeroMap/aeroPerformanceMap"

            self.tixi.uIDSetToXPath(get_xpath_parent(self.xpath), self.uid)
            create_branch(self.tixi, self.xpath)

        # Create and fill parameters fields
        for param in PARAMS:
            if param in self.df:
                if not self.df[param].isnull().values.any():
                    param_xpath = self.xpath + "/" + param
                    create_branch(self.tixi, param_xpath)
                    add_float_vector(self.tixi, param_xpath, self.df[param].tolist())
                else:
                    raise ValueError(
                        "All the 4 parameters (alt,mach,aos,aoa) must not contains NaN value to \
                            be saved in an aeroMap!"
                    )
            else:
                raise ValueError(
                    "All the 4 parameters (alt,mach,aos,aoa) must be define to save an aeroMap!"
                )

        # Create and fill coefficients fields
        for coef in COEFS:
            if coef in self.df:
                if not self.df[coef].isnull().values.all():
                    coef_xpath = self.xpath + "/" + coef
                    create_branch(self.tixi, coef_xpath)
                    add_float_vector(self.tixi, coef_xpath, self.df[coef].tolist())
                else:
                    print(
                        f'Warning: {coef} coefficient from "{self.uid}" aeroMap will not be \
                          written in the CPACS file because it contains only NaN!'
                    )

        # Create and fill damping derivatives fields
        for rates in ["negativeRates", "positiveRates"]:
            for damping_coef in DAMPING_COEFS:
                col_name = "dampingDerivatives_" + rates + "_" + damping_coef
                if col_name in self.df:
                    if not self.df[col_name].isnull().values.all():
                        coef_xpath = self.xpath + f"/dampingDerivatives/{rates}/{damping_coef}"
                        create_branch(self.tixi, coef_xpath)
                        add_float_vector(self.tixi, coef_xpath, self.df[col_name].tolist())
                    else:
                        print(
                            f'Warning: {damping_coef} coefficient from "{self.uid}" aeroMap will \
                              not be written in the CPACS file because it contains onlz NaN!'
                        )

        # Create and fill incrementMap fields
        # TODO

        # Create and fill the '/name' field
        name_xpath = get_xpath_parent(self.xpath) + "/name"
        create_branch(self.tixi, name_xpath)
        self.tixi.updateTextElement(name_xpath, self.name)

        # Create and fill the '/description' field
        description_xpath = get_xpath_parent(self.xpath) + "/description"
        create_branch(self.tixi, description_xpath)
        self.tixi.updateTextElement(description_xpath, self.description)

        # Create and fill the '/atmosphericModel' field
        atm_model_xpath = get_xpath_parent(self.xpath) + "/boundaryConditions/atmosphericModel"
        if not self.tixi.checkElement(atm_model_xpath):
            create_branch(self.tixi, atm_model_xpath)
        self.tixi.updateTextElement(atm_model_xpath, self.atmospheric_model)

    def export_csv(self, csv_path):
        """Export the AeroMap as a CSV file."""

        self.df.to_csv(csv_path, na_rep="NaN", index=False, float_format="%g")

    def get_cd0_oswald(self, ar, alt=None, mach=None, aos=None, plot=False):
        """Calculate and return CD0 and Oswald factor."""

        # Check for unique angleOfAttack condition
        aoa = self.get("angleOfAttack", alt, mach, aos)
        aoa_unique = list(set(aoa))

        if len(aoa) > len(aoa_unique):
            raise ValueError(
                "You must have unique angle of attack value to calculate CD0 and Oswald factor!"
            )

        # Get coefficient
        cd = self.get("cd", alt, mach, aos)
        cl = self.get("cl", alt, mach, aos)
        cl2 = np.power(cl, 2)

        # Remove value when Cl < 0
        cond = cl >= 0
        cl2 = cl2[cond]
        cd = cd[cond]
        cl = cl[cond]

        # Linear regression
        coef = np.polyfit(cl2, cd, 1)
        f = np.poly1d(coef)
        x = np.arange(0, 2, 0.01)
        y = f(x)

        # Calculate CD0 and Oswald factor e
        k, cd0 = coef
        e = 1 / (k * ar * math.pi)

        print("---------------------------------------------")
        print(f"For alt={alt}m, Mach={mach}, AoS={aos}deg:")
        print("CD0:", round(cd0, 6))
        print("Oswald factor:", round(e, 4))
        print("---------------------------------------------")

        if plot:
            _, ax = plt.subplots()
            ax.plot(cl2, cd, "o")
            ax.plot(x, y, "-")
            plt.show()

        return cd0, e

    def calculate_forces(self, aircraft):
        """Calculate forces and moment from coefficients"""

        COEF2FORCE_DICT = {"cd": "drag", "cl": "lift", "cs": "side"}
        COEF2MOMENT_DICT = {"cmd": "md", "cml": "ml", "cms": "ms"}

        def coef2force(row, coef):
            """Calculate force from coefficient"""

            dens = Atmosphere(row["altitude"]).density[0]
            sos = Atmosphere(row["altitude"]).speed_of_sound[0]

            return 0.5 * dens * aircraft.ref_area * (row["machNumber"] * sos) ** 2 * row[coef]

        for coef in COEF2FORCE_DICT:
            if coef in self.df:
                self.df[COEF2FORCE_DICT[coef]] = self.df.apply(
                    lambda row: coef2force(row, coef), axis=1
                )
            else:
                print(
                    f"Warning: {COEF2FORCE_DICT[coef]} will not be calculated because there is \
                    no {coef} coefficient in the aeroMap!"
                )

        for coef in COEF2MOMENT_DICT:
            if coef in self.df:
                self.df[COEF2MOMENT_DICT[coef]] = self.df.apply(
                    lambda row: coef2force(row, coef) * aircraft.ref_length, axis=1
                )
            else:
                print(
                    f"Warning: {COEF2MOMENT_DICT[coef]} will not be calculated because there is \
                    no {coef} coefficient in the aeroMap!"
                )

    def check_longitudinal_stability(self, alt=None, mach=None, aos=None):
        """Check longitudinal stability (cms vs aoa) with other parameters as filter (optional).
        The stability is checked by making a linear regression on cms vs aoa and analyzing the
        slope sign.

        Args:
            alt (list, optional): List of altitudes to filter. Defaults to None.
            mach (list, optional): List of Mach numbers to filter. Defaults to None.
            aos (list, optional): List of angle of sideslip to filter. Defaults to None.

        Return:
            stable (bool): Return if this set of conditions is stable (return False if neutral)
            msg (str): Message description

        """

        msg = ""

        alt_list = listify(alt)
        mach_list = listify(mach)
        aos_list = listify(aos)
        aoa_list = []

        if len(alt_list) > 1 or len(mach_list) > 1 or len(aos_list) > 1:
            msg = MSG_STAB_ONE_PARAM

        filt = get_filter(self.df, alt_list, mach_list, aos_list, aoa_list)
        df_filt = self.df.loc[filt]
        x = df_filt["angleOfAttack"].to_numpy()
        y = df_filt["cms"].to_numpy()

        if len(set(x)) < 2:
            return None, MSG_STAB_NOT_ENOUGH

        res = stats.linregress(x, y)

        if res.slope == 0:
            return False, MSG_STAB_NEUTRAL
        elif res.slope < 0:
            return True, msg
        else:
            return False, msg

    def check_directional_stability(self, alt=None, mach=None, aoa=None):
        """Check directional stability (cml vs aos) with other parameters as filter (optional).
        The stability is checked by making a linear regression on cml vs aos and analyzing the
        slope sign. With the CPACS angle convention cml vs aos slope must be positive to be stable.

        Args:
            alt (list, optional): List of altitudes to filter. Defaults to None.
            mach (list, optional): List of Mach numbers to filter. Defaults to None.
            aoa (list, optional): List of angle of attack to filter. Defaults to None.

        Return:
            stable (bool): Return if this set of conditions is stable (return False if neutral)
            msg (str): Message description

        """

        msg = ""

        alt_list = listify(alt)
        mach_list = listify(mach)
        aoa_list = listify(aoa)
        aos_list = []

        if len(alt_list) > 1 or len(mach_list) > 1 or len(aoa_list) > 1:
            msg = MSG_STAB_ONE_PARAM

        print(self.df)
        filt = get_filter(self.df, alt_list, mach_list, aos_list, aoa_list)
        df_filt = self.df.loc[filt]
        print(df_filt)
        x = df_filt["angleOfSideslip"].to_numpy()
        y = df_filt["cml"].to_numpy()

        print(x)

        if len(set(x)) < 2:
            return None, MSG_STAB_NOT_ENOUGH

        res = stats.linregress(x, y)

        if res.slope == 0:
            return False, MSG_STAB_NEUTRAL
        elif res.slope > 0:
            return True, msg
        else:
            return False, msg

    def check_lateral_stability(self, alt=None, mach=None, aoa=None):
        """Check lateral stability (cmd vs aos) with other parameters as filter (optional).
        The stability is checked by making a linear regression on cmd vs aos and analyzing the
        slope sign.

        Args:
            alt (list, optional): List of altitudes to filter. Defaults to None.
            mach (list, optional): List of Mach numbers to filter. Defaults to None.
            aoa (list, optional): List of angle of attack to filter. Defaults to None.

        Return:
            stable (bool): Return if this set of conditions is stable (return False if neutral)
            msg (str): Message description

        """

        msg = ""

        alt_list = listify(alt)
        mach_list = listify(mach)
        aoa_list = listify(aoa)
        aos_list = []

        if len(alt_list) > 1 or len(mach_list) > 1 or len(aoa_list) > 1:
            msg = MSG_STAB_ONE_PARAM

        filt = get_filter(self.df, alt_list, mach_list, aos_list, aoa_list)
        df_filt = self.df.loc[filt]
        x = df_filt["angleOfSideslip"].to_numpy()
        y = df_filt["cmd"].to_numpy()

        if len(set(x)) < 2:
            return None, MSG_STAB_NOT_ENOUGH

        res = stats.linregress(x, y)

        if res.slope == 0:
            return False, MSG_STAB_NEUTRAL
        elif res.slope < 0:
            return True, msg
        else:
            return False, msg

    def __str__(self):

        text_line = []
        text_line.append("\nAeroMap Description -------------------------------------------------")
        text_line.append(f"\nAeroMap uid: {self.uid}")
        text_line.append(f"AeroMap xpath: {self.xpath}")
        text_line.append(f"AeroMap description: {self.description}")
        text_line.append(f"Atmospheric model: {self.atmospheric_model}\n")
        text_line.append(f'Number of states: \t\t{self.df["angleOfAttack"].size}')
        alt_list = self.df["altitude"].unique().tolist()
        mach_list = self.df["machNumber"].unique().tolist()
        aos_list = self.df["angleOfSideslip"].unique().tolist()
        aoa_list = self.df["angleOfAttack"].unique().tolist()
        text_line.append(f'Unique altitude: \t\t{str(alt_list).strip("[]")}')
        text_line.append(f'Unique Mach number: \t\t{str(mach_list).strip("[]")}')
        text_line.append(f'Unique angle of sideslip:  \t{str(aos_list).strip("[]")}')
        text_line.append(f'Unique angle of attack: \t{str(aoa_list).strip("[]")}\n')
        text_line.append("Parameters and coefficients:")
        text_line.append(str(self.df))
        text_line.append("---------------------------------------------------------------------\n")

        return ("\n").join(text_line)



"""
!/usr/bin/env python3
-*- coding: utf-8 -*-

----------------------------------------------------------------------
Copyright 2021 CFS Engineering

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
----------------------------------------------------------------------

Author: Aidan Jungo

"""

from cpacspy.cpacsfunctions import get_tigl_configuration, get_value_or_default
from cpacspy.utils import AIRCRAFT_XPATH


class Aircraft:
    """Aircraft class"""

    def __init__(self, tixi, tigl):
        """Aircraft class to store references values and other information about the aircraft

        Args:
            tixi (object): TIXI object open from the CPACS file
            tigl (object): TIGL object open from the CPACS file
        """

        self.tixi = tixi
        self.tigl = tigl

        # Reference values
        reference_xpath = AIRCRAFT_XPATH + "/reference"
        self.ref_length = get_value_or_default(self.tixi, reference_xpath + "/length", 1)
        self.ref_area = get_value_or_default(self.tixi, reference_xpath + "/area", 1)
        self.ref_point_x = get_value_or_default(self.tixi, reference_xpath + "/point/x", 0)
        self.ref_point_y = get_value_or_default(self.tixi, reference_xpath + "/point/y", 0)
        self.ref_point_z = get_value_or_default(self.tixi, reference_xpath + "/point/z", 0)

        # Aircraft specific values (extract with TiGL)
        self.configuration = get_tigl_configuration(self.tigl)
        self.ref_wing_idx = self.get_main_wing_idx()  # By default reference wing is the largest

    @property
    def ref_wing_idx(self):
        return self._ref_wing_idx

    @ref_wing_idx.setter
    def ref_wing_idx(self, new_idx):
        self._ref_wing_idx = new_idx

        self._ref_wing_uid = self.configuration.get_wing(self._ref_wing_idx).get_uid()

        sym = 1
        if self.configuration.get_wing(self._ref_wing_idx).get_symmetry():
            sym = 2

        self.wing_span = self.configuration.get_wing(self._ref_wing_idx).get_wing_half_span() * sym
        self.wing_area = self.configuration.get_wing(self._ref_wing_idx).get_surface_area()
        self.wing_ar = self.configuration.get_wing(self._ref_wing_idx).get_aspect_ratio()

    @property
    def ref_wing_uid(self):
        return self._ref_wing_uid

    @ref_wing_uid.setter
    def ref_wing_uid(self, uid):
        self._ref_wing_uid = uid

        self._ref_wing_idx = self.configuration.get_wing_index(uid)

        sym = 1
        if self.configuration.get_wing(self._ref_wing_idx).get_symmetry():
            sym = 2

        self.wing_span = self.configuration.get_wing(self._ref_wing_uid).get_wing_half_span() * sym
        self.wing_area = self.configuration.get_wing(self._ref_wing_uid).get_surface_area()
        self.wing_ar = self.configuration.get_wing(self._ref_wing_uid).get_aspect_ratio()

    def get_main_wing_idx(self):
        """Find the largest wing index

        Args:
            self (object)
        """

        wing_area_max = 0
        wing_idx = None

        for i_wing in range(self.configuration.get_wing_count()):
            wing_area = self.configuration.get_wing(i_wing + 1).get_surface_area()

            if wing_area > wing_area_max:
                wing_area_max = wing_area
                wing_idx = i_wing + 1

        return wing_idx

    def __str__(self):

        text_line = []
        text_line.append("\nAircraft data -------------------------------------------------------")
        text_line.append(" ")
        text_line.append(f"Reference length: \t{self.ref_length} [m]")
        text_line.append(f"Reference area: \t{self.ref_area} [m^2]")
        text_line.append(
            f"Reference point: \t({self.ref_point_x},{self.ref_point_y},{self.ref_point_z}) [m]"
        )
        text_line.append(" ")
        text_line.append(f"Reference wing index: \t{self._ref_wing_idx}")
        text_line.append(f"Wing span: \t\t{self.wing_span} [m]")
        text_line.append(f"Wing area: \t\t{self.wing_area} [m^2]")
        text_line.append(f"Wing AR: \t\t{self.wing_ar} [-]")
        text_line.append(" ")
        text_line.append("---------------------------------------------------------------------\n")
        return ("\n").join(text_line)



"""
!/usr/bin/env python3
-*- coding: utf-8 -*-

----------------------------------------------------------------------
Copyright 2021 CFS Engineering

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
----------------------------------------------------------------------

Author: Aidan Jungo

"""

from cpacspy.cpacsfunctions import get_tigl_configuration, get_value_or_default

from cpacspy.utils import ROTORCRAFT_XPATH


class Rotorcraft:
    """Rotorcraft class"""

    def __init__(self, tixi, tigl):
        """Rotorcraft class to store references values and other information about the rotorcraf

        Args:
            tixi (object): TIXI object open from the CPACS file
            tigl (object): TIGL object open from the CPACS file
        """

        self.tixi = tixi
        self.tigl = tigl

        # Reference values
        reference_xpath = ROTORCRAFT_XPATH + "/reference"
        self.ref_length = get_value_or_default(self.tixi, reference_xpath + "/length", 1)
        self.ref_area = get_value_or_default(self.tixi, reference_xpath + "/area", 1)
        self.ref_point_x = get_value_or_default(self.tixi, reference_xpath + "/point/x", 0)
        self.ref_point_y = get_value_or_default(self.tixi, reference_xpath + "/point/y", 0)
        self.ref_point_z = get_value_or_default(self.tixi, reference_xpath + "/point/z", 0)

        # Rotorcraft specific values (extract with TiGL)
        self.configuration = get_tigl_configuration(self.tigl)
        self.rotor_count = self.configuration.get_rotor_count()

    def __str__(self):

        text_line = []
        text_line.append(
            "\nRotorcraft data -------------------------------------------------------"
        )
        text_line.append(" ")
        text_line.append(f"Reference length: \t{self.ref_length} [m]")
        text_line.append(f"Reference area: \t{self.ref_area} [m^2]")
        text_line.append(
            f"Reference point: \t({self.ref_point_x},{self.ref_point_y},{self.ref_point_z}) [m]"
        )
        text_line.append(" ")
        # text_line.append(f"Reference wing index: \t{self._ref_rotor_idx}")
        text_line.append(f"Rotor count: \t\t{self.rotor_count}")
        text_line.append(" ")
        text_line.append("---------------------------------------------------------------------\n")
        return ("\n").join(text_line)



"""
!/usr/bin/env python3
-*- coding: utf-8 -*-

----------------------------------------------------------------------
Copyright 2021 CFS Engineering

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
----------------------------------------------------------------------

Author: Aidan Jungo

"""

from pathlib import Path

import numpy as np

from cpacspy.utils import AIRCRAFT_XPATH, ROTORCRAFT_XPATH

try:
    from tixi3 import tixi3wrapper
    from tixi3.tixi3wrapper import Tixi3Exception
except ImportError:
    TIXI_INSTALLED = False
else:
    TIXI_INSTALLED = True

try:
    import tigl3.configuration
    from tigl3 import tigl3wrapper

    # from tigl3.tigl3wrapper import Tigl3Exception
except ImportError:
    TIGL_INSTALLED = False
else:
    TIGL_INSTALLED = True


def open_tixi(cpacs_path):
    """Create the TIXI Handle of a CPACS file given as input
    by its path. If this operation is not possible, it returns 'None'

    Source :
        * TIXI functions: http://tixi.sourceforge.net/Doc/index.html

    Args:
        cpacs_path (str): Path to the CPACS file

    Returns::
        tixi_handle (handles): TIXI Handle of the CPACS file
    """

    if not TIXI_INSTALLED:
        err_msg = """
        Unable to import Tixi. Please make sure Tixi is accessible to Python.
        Please refer to the documentation to check supported versions of Tixi.
        """
        print(err_msg)
        raise ModuleNotFoundError(err_msg)

    # To accept either a Path or a string
    if isinstance(cpacs_path, Path):
        cpacs_path = str(cpacs_path)

    tixi_handle = tixi3wrapper.Tixi3()
    tixi_handle.open(cpacs_path)

    print(f"TIXI handle has been created for {cpacs_path}.")

    return tixi_handle


def open_tigl(tixi_handle, rotorcraft=False):
    """Function 'open_tigl' return the TIGL Handle from its TIXI Handle.
    If this operation is not possible, it returns 'None'

    Source :
        * TIGL functions http://tigl.sourceforge.net/Doc/index.html

    Args:
        tixi_handle (handles): TIXI Handle of the CPACS file
        rotorcraft (bool, optional): Define if the aircraft is a rotorcraft.

    Returns:
        tigl_handle (handles): TIGL Handle of the CPACS file
    """

    if not TIGL_INSTALLED:
        err_msg = """
        Unable to import Tigl. Please make sure Tigl is accessible to Python.
        Please refer to the documentation to check supported versions of Tigl.
        """
        print(err_msg)
        raise ModuleNotFoundError(err_msg)

    # Get model uid to open TiGL handle (in case there is also a rotorcraft in the CPACS file)
    if rotorcraft:
        model_xpath = ROTORCRAFT_XPATH
    else:
        model_xpath = AIRCRAFT_XPATH

    if tixi_handle.checkAttribute(model_xpath, "uID"):
        model_uid = tixi_handle.getTextAttribute(model_xpath, "uID")
    else:
        model_uid = ""

    tigl_handle = tigl3wrapper.Tigl3()
    tigl_handle.open(tixi_handle, model_uid)

    tigl_handle.logSetVerbosity(1)  # 1 - only error, 2 - error and warnings

    return tigl_handle


def get_tigl_configuration(tigl):
    """Get the TiGL aircraft configuration manager."""

    # Get the configuration manager
    mgr = tigl3.configuration.CCPACSConfigurationManager_get_instance()
    aircraft = mgr.get_configuration(tigl._handle.value)

    return aircraft


def copy_branch(tixi, xpath_from, xpath_to):
    """Function to copy a CPACS branch.

    Function 'copy_branch' copy the branch (with sub-branches) from
    'xpath_from' to 'xpath_to' by using recursion. The new branch should
    be identical (uiD, attribute, etc).

    Source :
        * TIXI functions: http://tixi.sourceforge.net/Doc/index.html

    Args:
        tixi_handle (handles): TIXI Handle of the CPACS file
        xpath_from (str): xpath of the branch to copy
        xpath_to (str): Destination xpath

    Returns:
        tixi (handles): Modified TIXI Handle (with copied branch)
    """

    if not tixi.checkElement(xpath_from):
        raise ValueError(xpath_from + " XPath does not exist!")
    if not tixi.checkElement(xpath_to):
        raise ValueError(xpath_to + " XPath does not exist!")

    child_nb = tixi.getNumberOfChilds(xpath_from)

    if child_nb:

        child_list = []
        for i in range(child_nb):
            child_list.append(tixi.getChildNodeName(xpath_from, i + 1))

        # If it is a text Element --> no child
        if "#" in child_list[0]:
            elem_to_copy = tixi.getTextElement(xpath_from)
            tixi.updateTextElement(xpath_to, elem_to_copy)

        else:
            # If child are named child (e.g. wings/wing)
            if all(x == child_list[0] for x in child_list):
                namedchild_nb = tixi.getNamedChildrenCount(xpath_from, child_list[0])

                for i in range(namedchild_nb):
                    new_xpath_from = xpath_from + "/" + child_list[0] + "[" + str(i + 1) + "]"
                    new_xpath_to = xpath_to + "/" + child_list[0] + "[" + str(i + 1) + "]"
                    tixi.createElement(xpath_to, child_list[0])

                    # Call the function itself for recursion
                    copy_branch(tixi, new_xpath_from, new_xpath_to)

            else:
                for child in child_list:
                    new_xpath_from = xpath_from + "/" + child
                    new_xpath_to = xpath_to + "/" + child

                    # Create child
                    tixi.createElement(xpath_to, child)

                    # Call the function itself for recursion
                    copy_branch(tixi, new_xpath_from, new_xpath_to)

        # Copy attribute(s) if exists
        last_attrib = 0
        attrib_index = 1
        while not last_attrib:
            try:
                attrib_name = tixi.getAttributeName(xpath_from, attrib_index)
                attrib_text = tixi.getTextAttribute(xpath_from, attrib_name)
                tixi.addTextAttribute(xpath_to, attrib_name, attrib_text)
                attrib_index = attrib_index + 1
            except Tixi3Exception:
                last_attrib = 1


def get_uid(tixi, xpath):
    """Function to get uID from a specific XPath.

    Function 'get_uid' checks the xpath and get the corresponding uID.

    Source :
        * TIXI functions: http://tixi.sourceforge.net/Doc/index.html

    Args:
        tixi (handles): TIXI Handle of the CPACS file
        xpath (str): xpath of the branch to add the uid

    Returns:
        uid (str): uid to add at xpath
    """

    if not tixi.checkElement(xpath):
        raise ValueError(xpath + " XPath does not exist!")

    if tixi.checkAttribute(xpath, "uID"):
        uid = tixi.getTextAttribute(xpath, "uID")
        return uid
    else:
        raise ValueError("No uID found for: " + xpath)


def add_uid(tixi, xpath, uid):
    """Function to add UID at a specific XPath.

    Function 'add_uid' checks and add UID to a specific path, the function will
    automatically update the chosen UID if it exists already.

    Source :
        * TIXI functions: http://tixi.sourceforge.net/Doc/index.html

    Args:
        tixi (handles): TIXI Handle of the CPACS file
        xpath (str): xpath of the branch to add the uid
        uid (str): uid to add at xpath
    """

    exist = True
    uid_new = uid
    i = 0
    while exist is True:
        if not tixi.uIDCheckExists(uid_new):
            tixi.uIDSetToXPath(xpath, uid_new)
            exist = False
        else:
            i = i + 1
            uid_new = uid + str(i)
            print("UID already existing changed to: " + uid_new)


def add_value(tixi, xpath, value):
    """Add a value (string, integer of float) at the given XPath,
    if the node does not exist, it will be created. Values will be
    overwritten if paths exists.

    Args:
        tixi (handle): Tixi handle
        xpath (str): XPath of the vector to add
        value (list, tuple): Value to add
    """

    # Strip trailing '/' (has no meaning here)
    xpath = xpath.rstrip("/")

    # Get the field name and the parent CPACS path
    xpath_child_name = xpath.split("/")[-1]
    xpath_parent = xpath[: -(len(xpath_child_name) + 1)]

    if not tixi.checkElement(xpath_parent):
        create_branch(tixi, xpath_parent)

    if not tixi.checkElement(xpath):
        tixi.createElement(xpath_parent, xpath_child_name)

    tixi.updateTextElement(xpath, str(value))


def get_value(tixi, xpath):
    """Check first if the the xpath exist and that a value is stored
    at this place. Returns this value. It returns a:
    - float value if the value can be read as a float
    - boolean if the value is 'True'/'False',
    - otherwise a string

    Source :
        * TIXI functions: http://tixi.sourceforge.net/Doc/index.html

    Args:
        tixi_handle (handles): TIXI Handle of the CPACS file
        xpath (str): xpath of the value to get

    Returns:
         value (float, bool, str): Value found at xpath
    """

    if not tixi.checkElement(xpath):
        raise ValueError(f"{xpath} cannot be found in the CPACS file")

    value = tixi.getTextElement(xpath)

    if not value:
        raise ValueError(f"No value has been found at {xpath}")

    # Check if the value should be return as boolean
    if value == "True":
        return True
    elif value == "False":
        return False

    # Check if the value should be return as float
    try:
        float(value)
        return float(value)
    except ValueError:
        pass

    # Otherwise, return the value as a string
    return value


def get_value_or_default(tixi, xpath, default_value):
    """Do the same than the function 'get_value' but if no value is found
    at the xpath it returns the default value and add it in the CPACS file
    at the xpath. If the xpath does not exist, it is created.

    Source :
        * TIXI functions: http://tixi.sourceforge.net/Doc/index.html

    Args:
        tixi (handles): TIXI Handle of the CPACS file
        xpath (str): xpath of the value to get
        default_value (str, bool, float or int): Default value

    Returns:
        tixi (handles): Modified TIXI Handle (with added default value)
        default_value (str, bool, float or int): Value found at xpath
    """

    # Try to get the value in the CPACS file
    try:
        return get_value(tixi, xpath)
    except ValueError:
        pass

    # Prepare to write the default value in the CPACS file
    xpath_parent = get_xpath_parent(xpath, 1)
    create_branch(tixi, xpath_parent, False)
    value_name = xpath.split("/")[-1]

    is_bool = False
    is_float = False

    # Check if the default_value could be float
    try:
        value = float(default_value)
        is_float = True
    except ValueError:
        pass

    # Copy the default value (in case it is a string)
    if not is_float:
        value = default_value

    # Check if the value should be return as boolean
    if isinstance(default_value, bool):
        is_bool = True
        if default_value:
            value = True
        else:
            value = False

    # Write the default value in the CPACS file as a float
    if is_float and not is_bool:
        tixi.addDoubleElement(xpath_parent, value_name, value, "%g")
        return value

    # Write the default value in the CPACS file as a string (also for booleans)
    tixi.addTextElement(xpath_parent, value_name, str(value))

    # Return the value as a string
    return value


def get_float_vector(tixi, xpath):
    """Get a vector (composed by float) at the
    given XPath, if the node does not exist, an error will be raised.

    Args:
        tixi (handle): Tixi handle
        xpath (str): XPath of the vector to get
    """

    if not tixi.checkElement(xpath):
        raise ValueError(xpath + " path does not exist!")

    float_vector_str = tixi.getTextElement(xpath)

    if float_vector_str == "":
        raise ValueError("No value has been found at " + xpath)

    if float_vector_str.endswith(";"):
        float_vector_str = float_vector_str[:-1]

    float_vector_list = float_vector_str.split(";")
    float_vector = [
        np.nan if elem in ("nan", "NaN") else float(elem) for elem in float_vector_list
    ]

    return float_vector


def add_float_vector(tixi, xpath, vector):
    """Add a vector (composed by float) at the given XPath,
    if the node does not exist, it will be created. Values will be
    overwritten if paths exists.

    Args:
        tixi (handle): Tixi handle
        xpath (str): XPath of the vector to add
        vector (list, tuple): Vector of floats to add
    """

    # Strip trailing '/' (has no meaning here)
    if xpath.endswith("/"):
        xpath = xpath[:-1]

    # Get the field name and the parent CPACS path
    xpath_child_name = xpath.split("/")[-1]
    xpath_parent = xpath[: -(len(xpath_child_name) + 1)]

    if not tixi.checkElement(xpath_parent):
        create_branch(tixi, xpath_parent)

    vector = [float(v) for v in vector]

    if tixi.checkElement(xpath):
        tixi.updateFloatVector(xpath, vector, len(vector), format="%g")
        tixi.addTextAttribute(xpath, "mapType", "vector")
    else:
        tixi.addFloatVector(xpath_parent, xpath_child_name, vector, len(vector), format="%g")
        tixi.addTextAttribute(xpath, "mapType", "vector")


def add_string_vector(tixi, xpath, vector):
    """Add a vector (of string) at given CPACS xpath

    Function 'add_string_vector' will add a vector (composed by stings) at the
    given XPath, if the node does not exist, it will be created. Values will be
    overwritten if paths exists.

    Args:
        tixi (handle): Tixi handle
        xpath (str): XPath of the vector to add
        vector (list): Vector of string to add
    """

    # Strip trailing '/' (has no meaning here)
    if xpath.endswith("/"):
        xpath = xpath[:-1]

    # Get the field name and the parent CPACS path
    xpath_child_name = xpath.split("/")[-1]
    xpath_parent = xpath[: -(len(xpath_child_name) + 1)]

    vector_str = ";".join([str(elem) for elem in vector])

    if not tixi.checkElement(xpath_parent):
        create_branch(tixi, xpath_parent)

    if tixi.checkElement(xpath):
        tixi.updateTextElement(xpath, vector_str)
    else:
        tixi.addTextElement(xpath_parent, xpath_child_name, vector_str)


def get_string_vector(tixi, xpath):
    """Get a vector (of string) at given CPACS xpath

    Function 'get_string_vector' will get a vector (composed by string) at the
    given XPath, if the node does not exist, an error will be raised.

    Args:
        tixi (handle): Tixi handle
        xpath (str): XPath of the vector to get
    """

    if not tixi.checkElement(xpath):
        raise ValueError(xpath + " path does not exist!")

    string_vector_str = tixi.getTextElement(xpath)

    if string_vector_str == "":
        raise ValueError("No value has been fournd at " + xpath)

    if string_vector_str.endswith(";"):
        string_vector_str = string_vector_str[:-1]
    string_vector_list = string_vector_str.split(";")
    string_vector = [str(elem) for elem in string_vector_list]

    return string_vector


def get_xpath_parent(xpath, level=1):
    """Get the parent xpath at any level, 1 is parent just above the input xpath.

    Args:
        xpath (str): Input xpath
        level (int, optional): Parent level. Defaults to 1.

    Returns:
        str: Parent xpath
    """

    if not xpath.startswith("/"):
        raise ValueError('"get_xpath_parent" must recieve an xpath as argument!')

    if len(xpath.split("/")) - 1 <= level:
        raise ValueError("No parent available at this level")

    return "/".join(xpath.split("/")[:-level])


def create_branch(tixi, xpath, add_child=False):
    """Create a branch in the tixi handle and also all the missing parent nodes.
    Be careful, the xpath must be unique until the last element, it means,
    if several element exist, its index must be precised (index start at 1).
    e.g.: '/cpacs/vehicles/aircraft/model/wings/wing[2]/name'

    If the entire xpath already exist, the option 'add_child' (True/False) lets
    the user decide if a named child should be added next to the existing
    one(s). This only valid for the last element of the xpath.

    Source :
        * TIXI functions: http://tixi.sourceforge.net/Doc/index.html

    Args:
        tixi (handles): TIXI Handle of the CPACS file
        xpath (str): xpath of the branch to create
        add_child (boolean): Choice of adding a name child if the last element
                             of the xpath if one already exists

    Returns:
        tixi (handles): Modified TIXI Handle (with new branch)
    """

    xpath_split = xpath.split("/")
    xpath_count = len(xpath_split)

    for i in range(xpath_count - 1):
        xpath_index = i + 2
        xpath_partial = "/".join(str(m) for m in xpath_split[0:xpath_index])
        xpath_parent = "/".join(str(m) for m in xpath_split[0 : xpath_index - 1])
        child = xpath_split[(xpath_index - 1)]
        if tixi.checkElement(xpath_partial):
            if child == xpath_split[-1] and add_child:
                namedchild_nb = tixi.getNamedChildrenCount(xpath_parent, child)
                tixi.createElementAtIndex(xpath_parent, child, namedchild_nb + 1)
        else:
            tixi.createElement(xpath_parent, child)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import setuptools
from src.cpacspy.__version__ import __version__

NAME = "cpacspy"
VERSION = __version__
AUTHOR = "Aidan Jungo"
EMAIL = "aidan.jungo@cfse.ch"
DESCRIPTION = "PyPI package to work with CPACS file and AeroMaps"
LONG_DESCRIPTION = open("README.md").read()
URL = "https://github.com/cfsengineering/cpacspy"
REQUIRES_PYTHON = ">=3.7.0"
REQUIRED = []
README = "README.md"
PACKAGE_DIR = "src"
LICENSE = "Apache License 2.0"


setuptools.setup(
    name=NAME,
    version=VERSION,
    author=AUTHOR,
    author_email=EMAIL,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url=URL,
    include_package_data=True,
    package_dir={"": PACKAGE_DIR},
    license=LICENSE,
    packages=[NAME],
    python_requires=REQUIRES_PYTHON,
    keywords=["CPACS", "aircraft", "design", "xml", "aerodynamics", "coefficients", "databases"],
    install_requires=REQUIRED,
    # See: https://pypi.org/classifiers/
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
    ],
)
